########################################################################
# charset.py Version 0.4.1
# (c) by Michael Stroeder, michael.stroeder@propack-data.de
########################################################################
# Module for converting characters sets
########################################################################

import sys, string

# Alphabet for encrypted passwords (see module crypt)
crypt_alphabet = './0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
crypt_alphabet_len = len(crypt_alphabet)

known_charsets = ['ISO-8859-1','UTF-8','ASN.1','T.61','HTML4']

########################################################################
# Trys to escape all characters with a special meaning in HTML
# to appropriate character tags
########################################################################

def escapeHTML(s,escape_html_chars='<>":={}()'):
  if not s:
    return "" # seems to happen often
  if len(s) < 10:
    # this algorithm is faster for small strings
    def repl(c, esc=escape_html_chars):
      if c in esc:
        return '&#%d;' % ord(c)
      return c
    return string.join(map(repl, s), "")
  else:
    # this algorithm is faster for long strings
    for c in escape_html_chars:
      s=string.replace(s,c,'&#%d;'%ord(c))
    return s

########################################################################
# ISO-8859-1 in UTF-8 (Unicode) wandeln
########################################################################

def iso2utf(s):
  def repl(ch):
    if ch <= "\x7F":
      return ch
    else:
      c=ord(ch)
      return chr(0xC0 | (0x03 & (c >> 6)))+chr(0x80 | (0x3F & c))
  return string.join(map(repl, s), "")

UTF8len= ( 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	   1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	   2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 5, 6 )

UTF8mask = (0x3F,0x7F,0x1F,0x0F,0x07,0x03,0x01)

########################################################################
# UTF-8 (Unicode) nach ISO-8859-1 wandeln
# nicht in ISO-8859-1 darstellbare Zeichen werden zu '?'
########################################################################

def utf2iso(s):

  new = ''
  ind = 0
  slen = len(s)

  while ind < slen:
    c = s[ind]
    ind = ind+1
    if c <= "\x7F":
      # short circuit for ASCII
      new = new + c
      continue
    c=ord(c)
    clen = UTF8len[(c >> 2) & 0x3F]
    u = c & UTF8mask[clen]
    if clen==0:
      clen = 4
    else:
      clen = clen-1
    while clen and ind < slen:
      c=ord(s[ind])
      ind = ind+1
      if (c & 0xC0) == 0x80:
        u = (u << 6) | (c & 0x3F)
      else:
        ind = ind-1
        break
      clen = clen-1

    if (u <= 0xFF):
      new = new+chr(u)
    else:
      new = new+'?'

  return new

########################################################################
# UTF-8 (Unicode) in HTML 4 konforme Unicode-Darstellung wandeln
########################################################################

def utf2html4(s):

  new = ''
  ind = 0
  slen = len(s)

  while ind < slen:
    c = s[ind]
    ind = ind+1
    if c <= "\x7F":
      # short circuit for ASCII
      new = new + c
      continue
    c=ord(c)
    clen = UTF8len[(c >> 2) & 0x3F]
    u = c & UTF8mask[clen]
    if clen==0:
      clen = 4
    else:
      clen = clen-1
    while clen and ind < slen:
      c=ord(s[ind])
      ind = ind+1
      if (c & 0xC0) == 0x80:
        u = (u << 6) | (c & 0x3F)
      else:
        ind = ind-1
        break
      clen = clen-1

    if u<128:
      new = new + chr(u)
    else:
      new = new + '&#%d;' % u

  return new

########################################################################
# ISO-Zeichen nach HTML 4 wandeln
########################################################################

def iso2html4(s):
  def repl(ch):
    if ch <= "\x7F":
      return ch
    else:
      return '&#%d;' % ord(ch)
  return string.join(map(repl, s), "")

########################################################################
# ISO-8859-1 nach T61String wandeln
########################################################################

def iso2t61(s):
  def repl(ch):
    if ch <= "\x7F":
      return ch
    else:
      return '\\x%X' % ord(ch)
  return string.join(map(repl, s), "")

########################################################################
# T61String nach ISO-8859-1 wandeln
########################################################################

def t612iso(s, find=string.find, atoi=string.atoi, chr=chr):
  new = ''
  slashpos = find(s,'\\x')
  while slashpos!=-1:
    if (s[slashpos]==0) or (s[slashpos]>0 and s[slashpos-1]!='\\'):
      new = new+s[0:slashpos]+chr(atoi(s[slashpos+2:slashpos+4],16))
      s = s[slashpos+4:]
    else:
      new = new+s[0:slashpos-1]
      s = s[slashpos+1:]
    slashpos = find(s,'\\x')
  return new+s

########################################################################
# T61String nach HTML4 wandeln
########################################################################

def t612html4(s, find=string.find, atoi=string.atoi):
  new = ''
  slashpos = find(s,'\\x')
  while slashpos!=-1:
    if (s[slashpos]==0) or (s[slashpos]>0 and s[slashpos-1]!='\\'):
      new = new+s[0:slashpos]+'&#%d;' % atoi(s[slashpos+2:slashpos+4],16)
      s = s[slashpos+4:]
    else:
      new = new+s[0:slashpos-1]
      s = s[slashpos+1:]
    slashpos = find(s,'\\x')
  return new+s

########################################################################
# ISO-8859-1 nach ASN.1 wandeln
########################################################################

def iso2asn1(s):
  def repl(ch):
    if ch <= "\x7F":
      return '\\x00%s' % ch
    else:
      return '\\x00\\x%X%s' % (ord(ch), ch)
  return string.join(map(repl, s), "")

########################################################################
# ASN.1 nach ISO-8859-1 wandeln
########################################################################

def asn12iso(s):
  return t612iso(string.replace(s,'\\x00',''))

########################################################################
# ASN.1 nach HTML4 wandeln
########################################################################

def asn12html4(s):
  return t612html4(string.replace(s,'\\x00',''))

########################################################################
# Allgemeine Recoding-Routine
########################################################################

recode_func = {
  'ISO-8859-1': {
      'UTF-8'      : iso2utf,
      'HTML4'      : iso2html4
    },
  'UTF-8': {
      'HTML4'      : utf2html4,
      'ISO-8859-1' : utf2iso
    }
  }

def recode(s,source,target):

  if source==target:
    return s

  return recode_func[string.upper(source)][string.upper(target)](s)


def _testone(func, input, output):
  """ call "func" with "input" and compare its output to "output" """
  isout = func(input)
  if isout == output:
    print "%-10s OK" % func.__name__
  else:
    print func.__name__, "failed:"
    print "Input:   ", input
    print "Output:  ", isout
    print "Expected:", output

def _test():
  """ do some minimal tests """
  _testone(escapeHTML, "", "")
  _testone(escapeHTML, 'abcABC!\'"$%&/()=?<>[]',
          'abcABC!\'&#34;$%&/&#40;&#41;&#61;?&#60;&#62;[]')
  _testone(iso2utf, "", "")
  _testone(iso2utf, 'abcABC!\'"$%&/()=?<>[]',
          'abcABC\303\244\303\266\303\274\303\204\303\226\303\234\303\237'
          '!\'"\302\247$%&/()=?<>[]')
  _testone(utf2iso, "", "")
  _testone(utf2iso, 'abcABC\303\244\303\266\303\274\303\204\303\226\303\234'
          '\303\237!\'"\302\247$%&/()=?<>[]', 'abcABC!\'"$%&/()=?<>[]')
  _testone(utf2html4, "", "")
  _testone(utf2html4, 'abcABC\303\244\303\266\303\274\303\204\303\226\303\234'
          '\303\237!\'"\302\247$%&/()=?<>[]', 'abcABC&#228;&#246;&#252;'
          '&#196;&#214;&#220;&#223;!\'"&#167;$%&/()=?<>[]')
  _testone(iso2html4, "", "")
  _testone(iso2html4, 'abcABC!\'"$%&/()=?<>[]', 'abcABC&#228;&#246;'
          '&#252;&#196;&#214;&#220;&#223;!\'"&#167;$%&/()=?<>[]')
  _testone(iso2t61, "", "")
  _testone(iso2t61, 'abcABC!\'"$%&/()=?<>[]', 'abcABC\\xE4\\xF6'
          '\\xFC\\xC4\\xD6\\xDC\\xDF!\'"\\xA7$%&/()=?<>[]')
  _testone(t612iso, "", "")
  _testone(t612iso, 'abcABC\\xE4\\xF6\\xFC\\xC4\\xD6\\xDC\\xDF!\'"\\xA7'
          '$%&/()=?<>[]', 'abcABC!\'"$%&/()=?<>[]')
  _testone(t612html4, "", "")
  _testone(t612html4, 'abcABC\\xE4\\xF6\\xFC\\xC4\\xD6\\xDC\\xDF!\'"\\xA7'
          '$%&/()=?<>[]', 'abcABC&#228;&#246;&#252;&#196;&#214;&#220;'
          '&#223;!\'"&#167;$%&/()=?<>[]')
  _testone(iso2asn1, "", "")
  _testone(iso2asn1, 'abcABC!\'"$%&/()=?<>[]', '\\x00a\\x00b\\x00c'
          '\\x00A\\x00B\\x00C\\x00\\xE4\344\\x00\\xF6\366\\x00\\xFC\374'
          '\\x00\\xC4\304\\x00\\xD6\326\\x00\\xDC\334\\x00\\xDF\337\\x00!'
          '\\x00\'\\x00"\\x00\\xA7\247\\x00$\\x00%\\x00&\\x00/\\x00(\\x00)'
          '\\x00=\\x00?\\x00<\\x00>\\x00[\\x00]')
  _testone(asn12iso, "", "")
  _testone(asn12iso, '\\x00a\\x00b\\x00c\\x00A\\x00B\\x00C\\x00\\xE4\344'
          '\\x00\\xF6\366\\x00\\xFC\374\\x00\\xC4\304\\x00\\xD6\326\\x00'
          '\\xDC\334\\x00\\xDF\337\\x00!\\x00\'\\x00"\\x00\\xA7\247\\x00$'
          '\\x00%\\x00&\\x00/\\x00(\\x00)\\x00=\\x00?\\x00<\\x00>\\x00[\\x00]',
          'abcABC!\'"$%&/()=?<>[]')
  _testone(asn12html4, "", "")
  _testone(asn12html4, '\\x00a\\x00b\\x00c\\x00A\\x00B\\x00C\\x00\\xE4\344'
          '\\x00\\xF6\366\\x00\\xFC\374\\x00\\xC4\304\\x00\\xD6\326\\x00'
          '\\xDC\334\\x00\\xDF\337\\x00!\\x00\'\\x00"\\x00\\xA7\247\\x00$'
          '\\x00%\\x00&\\x00/\\x00(\\x00)\\x00=\\x00?\\x00<\\x00>\\x00[\\x00]',
          'abcABC&#228;&#246;&#252;&#196;&#214;&#220;&#223;!\'"&#167;$%&/()'
          '=?<>[]')

if __name__ == "__main__":
  _test()

